#include "local.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <alloca.h>
#include <errno.h>
#include <time.h>

#include <sys/mman.h>
#include <sys/stat.h>

/*

  Copyright (c) 2024 Juha Railoma

  Permission to use, copy, modify, and/or distribute this software for
  any purpose with or without fee is hereby granted.

  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
  WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
  OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
  FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
  OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

*/

/* statics */
static int             __link_files_open;
static int64_t         __link_errcount;
static const char     *__link_tmphash;

/* sorters copy deliberately pointers, not structs */
static int
__link_compare_cksum(const void *_a, const void *_b) {

    const struct z_compare_t **a = N_CAST(_a, a);
    const struct z_compare_t **b = N_CAST(_b, b);

    if ((*a)->file->nlink > (*b)->file->nlink) return -1;
    if ((*a)->file->nlink < (*b)->file->nlink) return 1;
    return memcmp((*a)->u.cksum.sha256, (*b)->u.cksum.sha256, 32);
}

static void
__link_sort_cksum(struct z_compare_t **clist, uint64_t count) {
    qsort(clist, count, n_asize0(clist),
          __link_compare_cksum);
}

static int
__link_compare_attrs(const void *_a, const void *_b) {

    const struct z_compare_t **a = N_CAST(_a, a);
    const struct z_compare_t **b = N_CAST(_b, b);

    if ((*a)->u.attrs.uid > (*b)->u.attrs.uid) return 1;
    if ((*a)->u.attrs.uid < (*b)->u.attrs.uid) return -1;
    if ((*a)->u.attrs.gid > (*b)->u.attrs.gid) return 1;
    if ((*a)->u.attrs.gid < (*b)->u.attrs.gid) return -1;
    if ((*a)->u.attrs.mode > (*b)->u.attrs.mode) return 1;
    if ((*a)->u.attrs.mode < (*b)->u.attrs.mode) return -1;
    return 0;
}

static void
__link_sort_attrs(struct z_compare_t **clist, uint64_t count) {
    qsort(clist, count, n_asize0(clist),
          __link_compare_attrs);
}

static int
__link_compare_bytes(const void *_a, const void *_b) {

    const struct z_compare_t **a = N_CAST(_a, a);
    const struct z_compare_t **b = N_CAST(_b, b);

    if ((*a)->u.bytes.mask > (*b)->u.bytes.mask) return 1;
    if ((*a)->u.bytes.mask < (*b)->u.bytes.mask) return -1;
    return 0;
}

static void
__link_sort_bytes(struct z_compare_t **clist, uint64_t count) {
    qsort(clist, count, n_asize0(clist),
          __link_compare_bytes);
}

/* utils */
static int
__link_open_ce(struct z_compare_t *ce) {

    const char *path;

    if (ce->u.bytes.fd >= 0) {

        /* file is already open, but caller expects it just opened
         * ie. seeked to position 0 */
        if (lseek(ce->u.bytes.fd, 0, SEEK_SET) < 0)
            return err_perror("  re-seek failed on: %s",
                              n_stackcat(ce->file->parent,
                                         ce->file->name));
        return 0;
    }

    path = n_stackcat(ce->file->parent, ce->file->name);
    if ((ce->u.bytes.fd = open(path, O_RDONLY)) < 0)
        return err_perror("  open file %s failed", path);

    __link_files_open++;
    return 0;
}

static int
__link_close_ce(struct z_compare_t *ce, int force) {

    /* be not very eager to close as long we are under safe file
     * count, we probably can avoid reopening */
    if (!force && (__link_files_open < app_fds_safe))
        return 0;

    if (ce->u.bytes.fd < 0) /* already close */
        return 0;

    if (close(ce->u.bytes.fd) < 0)
        return err_perror("  close file failed");

    /* success */
    ce->u.bytes.fd = -1;
    __link_files_open--;
    return 0;
}

static int
__link_mask_byte(struct z_compare_t *ce, uint32_t *mask,
                 off_t pos, int shift) {

    unsigned char buf[1];

    if (lseek(ce->u.bytes.fd, pos, SEEK_SET) < 0)
        return err_perror("  seek(mask) failed");
    if (read(ce->u.bytes.fd, buf, 1) != 1)
        return err_perror("  read(mask) failed");

    *mask |= ((uint32_t)buf[0]) << shift;
    return 0;
}

/* phases */
static int
__link_wrap(const char *srcpath,
            const char *tgtpath) {

    char *tmppath, *cpos;

    tmppath = cpos = n_stackcat(srcpath, __link_tmphash);
    cpos += strlen(cpos) - 2;

    /* try paths */
    for(char i = 'A'; i <= 'Z'; i++) {
        for(char j = 'A'; j <= 'Z'; j++) {

            cpos[0] = i;
            cpos[1] = j;

            if (!link(tgtpath, tmppath)) { /* success */

                int terrno;
                if (!rename(tmppath, srcpath)) {
                    log_verbose("  hardlinked %s to %s",
                                srcpath, tgtpath);
                    return 0; /* success */
                }

                /* else remove leftover tempfile */
                terrno = errno;
                if (unlink(tmppath) < 0) /* never */
                    return err_print("  rename link failed: %s\n"
                                     "     source: %s\n"
                                     "     target: %s\n"
                                     "  CANNOT REMOVE LEFTOVER TMPLINK!\n"
                                     "     %s: %s",
                                     strerror(terrno),
                                     srcpath,
                                     tgtpath,
                                     tmppath,
                                     strerror(errno));

                /* else replacing target failed, probably because on
                 * permissions or ownership. this is not normal. print
                 * error but continue */
                if (!app_quiet)
                    err_perror("  cannot replace target %s", tgtpath);
                return 0;
            }

            /* link-creation failed */
            if (errno == EEXIST) {

                /* link-tempname exists, it's ok but print since
                 * it might suggest leftover templink */
                log_print("  tempname %s exists", tmppath);
                continue;
            }

            /* otherwise not point of retrying and expectin different
             * result. anyway, don't stop going */
            if (!app_quiet)
                err_perror("  cannot replace %s", tgtpath);

            return -1;
        }
    }

    log_print("  no free name for temporary hardlink found.\n"
              "  the last tried was %s", tmppath);
    return -1;
}

static int
__link_link(struct z_compare_t *src, struct z_compare_t *tgt) {

    struct stat st;
    const char *srcpath;
    const char *tgtpath;

    if (tgt->file->nlink >= app_nlinks_limit) {
        return 1;
    }

    if (tgt->file->ino == src->file->ino) { /* same file */
        return 1;
    }

    /* trailing slash is always contained on parent */
    srcpath = n_stackcat(src->file->parent, src->file->name);
    tgtpath = n_stackcat(tgt->file->parent, tgt->file->name);

    if (app_dry_run) {

        tgt->file->nlink++;
        log_verbose("  would link %s to %s", srcpath, tgtpath);

        if (stat(tgtpath, &st) < 0) {
            err_print("    restat failed on %s", tgtpath);
        } else {

            /* update-stats */
            stat_linked_files++;
            stat_linked_bytes += st.st_size;
            stat_linked_blocks += st.st_blocks;
        }

        return 0;
    }

    if (__link_wrap(srcpath, tgtpath) < 0) {

        /* links might overflow despite trying to keep track. that's
         * ok, forward to next candidate */
        if (errno == EMLINK) {
            log_debug("  link count max reached: %s", tgtpath);
            return 1;
        }

        /* on fault-safe mode, ff to next potential target, link_wrap
         * has already printed error */
        if (app_keep_going) {
            __link_errcount++;
            return 1;
        }

        return -1;
    }

    /* restat to get actual link count, it's relatively cheap */
    if (stat(tgtpath, &st) < 0) {
        err_print("  restat failed on %s", tgtpath);
        tgt->file->nlink++;
    } else {

        /* update-stats */
        stat_linked_files++;
        stat_linked_bytes += st.st_size;
        stat_linked_blocks += st.st_blocks;

        /* hard-clamp max links */
        tgt->file->nlink = (st.st_nlink > 0x7fffffff) ?
            0x7fffffff : (st.st_nlink & 0x7fffffff);
    }

    return 0;
}

static int
__link_actual(struct z_compare_t **clist) {

    struct z_compare_t **cptr;

    /* clist is sorted by link-count descending. find best suitable
     * target to link */

restart:

    while(*clist && ((*clist)->file->nlink >= app_nlinks_limit)) {
        if (!(*(++clist))) {
            /* no suitable candidates left to link to */
            return 0;
        }
    }

    cptr = clist;
    if (!(*(++cptr))) {
        /* nothing left to link */
        return 0;
    }

    while(*cptr) {

        /* try hardlink */
        switch(__link_link(*cptr, *clist)) {
        case 0: /* next */
            cptr++;
            break;
        case 1: /* target-full, */
            clist = cptr;
            goto restart;
        default:
            return -1;
        }
    }

    return 0;
}

static int
__link_slice_cksum(struct z_compare_t **clist, size_t size) {

    uint64_t count = 1;
    struct z_compare_t *ctmp, **cptr = clist;

    /* clist is list of fingerprinted same size files. if finerprints
     * match, files needs to be checksummed to get actual equal
     * files */
    for(++cptr; *cptr; cptr++) {

        if (!memcmp((*cptr)->u.cksum.sha256,
                    (*clist)->u.cksum.sha256,
                    32)) {

            count++;
            continue; /* equals */
        }

        if (count < 2) { /* no equals found, restart */
            count = 1;
            clist = cptr;
            continue;
        }

        /* else at least two consecutives are equal, temporarily null
         * terminate slice (theres always extra entry on end) */
        ctmp = *cptr;
        *cptr = NULL;
        if (__link_actual(clist) < 0)
            return -1;

        *cptr = ctmp;
        clist = cptr;
        count = 1;
    }

    /* remainder */
    return (count > 1) ? __link_actual(clist) : 0;
}

static int
__link_by_cksum(struct z_compare_t **clist,
                uint64_t count, size_t size) {

    uint64_t i;
    char buffer[0x8000]; /* files not fitting buffer will be mmapped */

    if (size < sizeof(buffer)) { /* plain old read might be more
                                  * efficient for small files
                                  * ie. mmapping couple bytes just
                                  * feels stupid */

        for(i = 0; i < count; i++) {

            if (__link_open_ce(clist[i]) < 0)
                return -1;

            if (app_debug) {
                log_debug("  cksum(read): %s",
                          n_stackcat(clist[i]->file->parent,
                                     clist[i]->file->name));
            }

            if (read(clist[i]->u.cksum.fd, buffer, size) != size) {
                __link_close_ce(clist[i], 1);
                return err_perror("  read(cksum) failed");
            }

            if (__link_close_ce(clist[i], 1) < 0)
                return -1;

            /* checksumming */
            clist[i]->u.cksum.sha256 =
                sha256(obstack_alloc(obs, 32), buffer, size);

            if (app_verbose && i && !(i % 10000)) {
                log_verbose("   contents cksum %zi/%zi", i, count);
            }
        }

    } else { /* checksum through memorymapping */

        char *base;

        for(i = 0; i < count; i++) {

            /* checksum through memorymapping */
            if (__link_open_ce(clist[i]) < 0)
                return -1;

            if (app_debug) {
                log_debug("  cksum(mmap): %s",
                          n_stackcat(clist[i]->file->parent,
                                     clist[i]->file->name));
            }

            if ((base = mmap(NULL, size, PROT_READ, MAP_PRIVATE,
                             clist[i]->u.cksum.fd, 0)) == MAP_FAILED) {

                __link_close_ce(clist[i], 1);
                return err_perror("  mmap failed");
            }

            /* advice unnecessary cache polluting off and hint about
             * sequential read for chsumming */
#ifdef MADV_POPULATE_READ
            if (size > 0x10000)
                madvise(base, size, (MADV_SEQUENTIAL |
                                     MADV_POPULATE_READ));
#else
            if (size > 0x10000)
                madvise(base, size, (MADV_SEQUENTIAL | MADV_WILLNEED));
#endif
            /* checksumming */
            clist[i]->u.cksum.sha256 =
                sha256(obstack_alloc(obs, 32), base, size);

#ifdef MADV_FREE /* linux >4.5 */
            if (size > 0x10000)
                madvise(base, size, (MADV_FREE));
#endif

            /* cleanup */
            if (munmap(base, size) < 0) {
                __link_close_ce(clist[i], 1);
                return err_perror("  munmap failed");
            }

            /* can be closed final. orgiginal caller verifies anyway */
            if (__link_close_ce(clist[i], 1) < 0)
                return -1;

            if (app_verbose && i && !(i % 10000)) {
                log_verbose("   contents cksum %zi/%zi", i, count);
            }
        }
    }

    /* re-sort */
    __link_sort_cksum(clist, count);

    /* and slice by equal masks. these needs to be checksummed */
    if (__link_slice_cksum(clist, size) < 0)
        return -1;

    /* ready and success */
    return 0;
}

static int
__link_slice_bytes(struct z_compare_t **clist, size_t size) {

    uint64_t count = 1;
    struct z_compare_t *ctmp, **cptr = clist;

    /* clist is list lightly fingerprinted same size files. if
     * finerprints match, files needs to be sha256summed to get actual
     * equal files */
    for(++cptr; *cptr; cptr++) {

        void *freeptr; /* only for freeing shasums off obstack */

        if (((*cptr)->u.bytes.mask == (*clist)->u.bytes.mask)) {
            count++;
            continue; /* equals */
        }

        if (count < 2) { /* no equals found, restart */
            count = 1;
            clist = cptr;
            continue;
        }

        /* else at least two consecutives are equal, temporarily null
         * terminate slice (theres always extra entry on end) */
        ctmp = *cptr;
        *cptr = NULL;

        /* obstack grows with sha256s, revert afterwards */
        freeptr = obstack_alloc(obs, sizeof(freeptr));
        if (__link_by_cksum(clist, count, size) < 0)
            return -1;
        obstack_free(obs, freeptr);

        *cptr = ctmp;
        clist = cptr;
        count = 1;
    }

    /* remainder */
    return (count > 1) ? __link_by_cksum(clist, count, size) : 0;
}

static int
__link_by_bytes(struct z_compare_t **clist,
                uint64_t count, size_t size) {

    uint64_t i;
    off_t pos0 = 0, pos1 = 1, pos2 = 2, pos3 = (size - 1);

    /* spread bytes */
    if (size > 1048576) {
        pos1 = 524287;
        pos2 = 1048576;
    } else if (size > 131072) {
        pos1 = 65535;
        pos2 = 131072;
    } else if (size > 256) {
        pos1 = 127;
        pos2 = 256;
    } else if (size > 64) {
        pos1 = 31;
        pos2 = 64;
    } else if (size < 3) {
        pos1 = pos2 = 0;
    };

    for(i = 0; i < count; i++) {

        uint32_t mask = 0;

        /* this is the first time open */
        if (__link_open_ce(clist[i]) < 0)
            return -1;

        if ((__link_mask_byte(clist[i], &mask, pos0, 0) < 0) ||
            (__link_mask_byte(clist[i], &mask, pos1, 8) < 0) ||
            (__link_mask_byte(clist[i], &mask, pos2, 16) < 0) ||
            (__link_mask_byte(clist[i], &mask, pos3, 24) < 0)) {

            __link_close_ce(clist[i], 0);
            return -1;
        }

        if (__link_close_ce(clist[i], 0) < 0)
            return -1;

        clist[i]->u.bytes.mask = mask;

        /* progress */
        if (app_verbose && i && !(i % 10000)) {
            log_verbose("   contents precollect %zi/%zi", i, count);
        }
    }

    /* re-sort */
    __link_sort_bytes(clist, count);

    /* and slice by equal bytepoints. these needs to be checksummed */
    if (__link_slice_bytes(clist, size) < 0)
        return -1;

    /* ready and success */
    return 0;
}

static int
__link_slice_attrs(struct z_compare_t **clist, size_t size) {

    uint64_t count = 1;
    struct z_compare_t *ctmp, **cptr = clist;

    /* clist is list of same size files, sorted by order of other
     * meaningingful attributes. next thing is to pick slice of equal
     * entries and resort them by contents */
    for(++cptr; *cptr; cptr++) {

        if (((*cptr)->u.attrs.uid == (*clist)->u.attrs.uid) &&
            ((*cptr)->u.attrs.gid == (*clist)->u.attrs.gid) &&
            ((*cptr)->u.attrs.mode == (*clist)->u.attrs.mode)) {
            count++;
            continue; /* equals */
        }

        if (count < 2) { /* no equals found, restart */
            count = 1;
            clist = cptr;
            continue;
        }

        /* else at least two consecutives are equal, temporarily null
         * terminate slice (theres always extra entry on end) */
        ctmp = *cptr;
        *cptr = NULL;
        if (__link_by_bytes(clist, count, size) < 0)
            return -1;

        *cptr = ctmp;
        clist = cptr;
        count = 1;
    }

    /* remainder */
    return (count > 1) ? __link_by_bytes(clist, count, size) : 0;
}

static int
__link_files(struct z_sizes_t *sp) {

    uint64_t i;
    struct z_file_t *fe;
    struct z_compare_t *ce, *centries, **clist;

    clist = obstack_alloc(obs, n_asizec(clist, sp->count + 1));
    centries = obstack_alloc(obs, n_asizec(centries, sp->count));

    /* construct struct and duplicate variables for fastern and easier
     * access */
    if (app_content_only) {

        /* omit unnecessary */
        for(i = 0, fe = sp->files; fe; fe = fe->next) {
            ce = &centries[i];
            ce->u.bytes.fd = -1; /* init-to-unopened */
            ce->file = fe;
            clist[i++] = ce;
        }

    } else {

        for(i = 0, fe = sp->files; fe; fe = fe->next) {
            ce = &centries[i];
            ce->u.attrs.fd = -1; /* init-to-unopened */
            ce->u.attrs.mode = fe->mode;
            ce->u.attrs.uid = fe->uid;
            ce->u.attrs.gid = fe->gid;
            ce->file = fe;
            clist[i++] = ce;
        }
    }

    /* terminate */
    clist[sp->count] = NULL;

    /* hardlinking */
    if (app_content_only) {

        /* contents only, all entries are linkable shortcut to compare
         * only contents */
        if (__link_by_bytes(clist, sp->count, sp->size) < 0)
            goto fault;

    } else {

        /* uid, gid and mode matters too, size-list must be sub-sliced
         * by them */
        __link_sort_attrs(clist, sp->count);
        if (__link_slice_attrs(clist, sp->size) < 0)
            goto fault;
    }

    /* size done, close open files */
    for(i = 0; i < sp->count; i++)
        __link_close_ce(clist[i], 1);

    /* reset obstack to initial position */
    log_debug("   obstack memory used: %zi bytes",
              (obstack_alloc(obs, 1) - (void *)clist));

    obstack_free(obs, clist);
    return 0;

fault:
    obstack_free(obs, clist);
    return -1;
}

int
link_dirs() {

    int i;
    int64_t pfiles, pbytes, pblocks;
    int64_t files_done = 0, files_mark = 0;
    char tmphash[64];
    struct z_sizes_t *s;

    /* semirandom tag for tmpfiles */
    __link_tmphash = tmphash;
    snprintf(tmphash, sizeof(tmphash), ".zhardlink.%08x%08x.XX",
             (uint32_t)getpid(), (uint32_t)time(NULL));

    /* enumerate through all sizes, sort them and link matching */
    if (app_dry_run) {
        log_print("Enumerating and hardlinking... (--dry-run)");
    } else {
        log_print("Enumerating and hardlinking...");
    }

    for(i = 0; i < Z_SIZES_HASH; i++) {
        for(s = app_sizes[i]; s; s = s->next) {

            log_debug(" File size %zu has %zu entries",
                      s->size, (uint64_t)s->count);

            files_done += s->count;
            if (s->count < 2) {
                continue;
            }

            pfiles = stat_linked_files;
            pbytes = stat_linked_bytes;
            pblocks = stat_linked_blocks;
            if (s->count > 1 && (__link_files(s) < 0))
                goto breakout;

            /* progress */
            if (files_done > (files_mark + 10000)) {
                log_print(" Progress %zi/%zi   (compared/total)",
                          files_done,
                          stat_origin_files);
                files_mark = files_done;
            }

            log_debug("   %zi files linked, %zi/%zi bytes/blocks saved",
                      (stat_linked_files - pfiles),
                      (stat_linked_bytes - pbytes),
                      (stat_linked_blocks - pblocks));
        }
    }

breakout:

    if (i < Z_SIZES_HASH) {
        log_print("Terminated because of previous error");
    } else if (__link_errcount) {
        log_print("There were %zi errors in hardlinking", __link_errcount);
    }

    /* print totals */
    logn_print("Total %zi files, ", stat_linked_files);
    if (stat_linked_bytes > (16 * 1024 * 1024)) {
        logn_print("%zi MiB", (stat_linked_bytes / (1024 * 1024)));
    } else if (stat_linked_bytes > 8192) {
        logn_print("%zi KiB", (stat_linked_bytes / 1024));
    } else {
        logn_print("%zi bytes", stat_linked_bytes);
    }

    logn_print(" linked (");

    if (stat_linked_blocks > (2 * 16 * 1024)) {
        logn_print("%zi MiB", ((stat_linked_blocks * 512) / 1024 / 1024));
    } else if (stat_linked_blocks > (2 * 8192)) {
        logn_print("%zi KiB", ((stat_linked_blocks * 512) / 1024));
    } else {
        logn_print("%zi bytes",  (stat_linked_blocks * 512));
    }

    log_print(" disk space saved)");

    if (app_dry_run) {
        log_print("Because of dry run mode SIZES ARE ESTIMATES");
    }

    return 0;
}

/*
  Local variables: ***
  mode: c ***
  coding: iso-8859-15-unix ***
  c-basic-offset: 4 ***
  indent-tabs-mode: nil ***
  End: ***
*/
