#ifndef _LOCAL_H_
#define _LOCAL_H_

#include <stdint.h>
#include <alloca.h>
#include <obstack.h>
#include <sys/types.h>

/*

  Copyright (c) 2024 Juha Railoma

  Permission to use, copy, modify, and/or distribute this software for
  any purpose with or without fee is hereby granted.

  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
  WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
  OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
  FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
  OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

*/

/* config */
#define Z_SIZES_HASH      4096

/* obstack */
#define obstack_chunk_alloc malloc
#define obstack_chunk_free  free

#define obstack_calloc(o, s) (memset(obstack_alloc((o), (s)), 0, (s)))
#define obstack_strdup(o, s) (obstack_copy0((o), (s), strlen((s))))


/* debugger breakpoint */
#if (defined (__i386__) || defined (__x86_64__)) &&	\
    (defined (__GNUC__)) && (__GNUC__ >= 2)
# define __BREAK()   { __asm__ __volatile__ ("int $03"); }
#elif (defined (__arm__) || defined ((__aarch64__)) &&	\
    (defined (__GNUC__)) && (__GNUC__ >= 2)
# define __BREAK()   { __asm__ __volatile__ ("brk #0x1"); }
#else
# include <signal.h>
# define __BREAK()   { raise(SIGTRAP); }
#endif

/* macros */
#define n_sizeof0(x) (sizeof((x)) - 1)
#define n_asize(x)  (sizeof((x)) / sizeof((x)[0]))
#define n_asize0(x) (sizeof((x)[0]))
#define n_asizec(x, y)  (sizeof((x)[0]) * (y))
#define n_cfree(x)  { if ((x)) free((void *)(x)); (x) = NULL; }
#define n_clear(x)  (memset((x), 0, sizeof(*(x))))
#define n_stmalloc(x) (malloc(sizeof(*(x))))
#define n_stcalloc(x) (calloc(1, sizeof(*(x))))
#define n_calloca(x) (memset(alloca((x)), 0, (x)))
#define n_stmalloca(x) (alloca(sizeof(*(x))))
#define n_stcalloca(x) (memset(alloca(sizeof(*(x))), 0, sizeof(*(x))))
#define n_eoa(x) ((x)[sizeof(x) - 1])
#define n_nullterm(x) (((x)[sizeof(x) - 1] = 0), (x))
#define n_isnulltermed(x) ((x)[sizeof(x) - 1] == 0)
#define n_stackdup(x)                           \
    (strcpy(alloca(strlen(x) + 1), (x)))
#define n_stackdup_bin(x, l)			\
    (memcpy(alloca(l), (x), (l)))
#define n_stackcat(x, y)					\
    (strcat(strcpy(alloca(strlen(x) + strlen(y) + 2), (x)), (y)))
#define n_stackcat_3(x, y, z)                                           \
    (strcat(strcat(strcpy(alloca(strlen(x) +                            \
                                 strlen(y) +                            \
                                 strlen(z) + 1), (x)), (y)), (z)))
#define n_stackcat_4(x, y, z, a)                                \
    (strcat(strcat(strcat(strcpy(alloca(strlen(x) +             \
                                        strlen(y) +             \
                                        strlen(z) +             \
                                        strlen(a) + 1),         \
                                 (x)), (y)), (z)), (a)))
#define n_stackcat_5(x, y, z, a, b)                                     \
    (strcat(strcat(strcat(strcat(strcpy(alloca(strlen(x) +              \
                                               strlen(y) +              \
                                               strlen(z) +              \
                                               strlen(a) +              \
                                               strlen(b) + 1),          \
                                        (x)), (y)), (z)), (a)), (b)))
#define n_strlen0(x)				\
    (strlen(x) + 1)
#define N_CAST(x,z) (__typeof__ (z))(x)

/* types */
struct z_slist_t {
    void *val;
    struct z_slist_t *next;
};

struct z_sizes_t {

    off_t size;
    uint64_t count;

    struct z_file_t *files;
    struct z_sizes_t *next;
};

struct z_file_t {

    uid_t uid;
    gid_t gid;
    ino_t ino;
    mode_t mode;
    int32_t nlink;

    const char *parent;
    const char *name;
    struct z_file_t *next;
};

struct z_compare_t {

    /* union for space-saving, fd must be on same location for
     * all. sized discovered are processed one by one. z_compare_t is
     * obstack allocated for every iteration and rewinded back after
     * particular size is processed

     * if there is zilliard of same size files, they consume lot if
     * not compacted */
    union {
        struct {
            int fd;
            mode_t mode;
            uid_t uid;
            gid_t gid;
        } attrs;

        struct {
            int fd;
            uint32_t mask;
        } bytes;

        /* for that reason sha is also pointer to sum stored in
         * rewindable obstack */
        struct {
            int fd;
            const char *sha256;
        } cksum;
    } u;

    struct z_file_t *file;
};

/* common.c, statistics counters */
extern int64_t stat_blk_count;
extern int64_t stat_chr_count;
extern int64_t stat_dir_count;
extern int64_t stat_lnk_count;
extern int64_t stat_fifo_count;
extern int64_t stat_file_count;
extern int64_t stat_sock_count;
extern int64_t stat_unknown_count;
extern int64_t stat_dir_open_failed;
extern int64_t stat_dir_stat_failed;
extern int64_t stat_file_stat_failed;

extern int64_t stat_origin_files;
extern int64_t stat_origin_bytes;

extern int64_t stat_linked_files;
extern int64_t stat_linked_bytes;
extern int64_t stat_linked_blocks;

/* util.c, shared variables */
extern struct obstack     *obs;
extern const char         *app_progname;

extern int                 app_quiet;
extern int                 app_debug;
extern int                 app_verbose;

extern int                 app_dry_run;
extern int                 app_keep_going;
extern int                 app_content_only;

extern int                 app_fds_max;
extern int                 app_fds_safe;
extern int                 app_path_max;
extern int                 app_nlinks_max;
extern int                 app_nlinks_limit;

extern int                 app_dirs_count;
extern char              **app_dirs;

extern int                 app_sizes_count;
extern struct z_sizes_t   *app_sizes[];

/* funcs */
extern const char *sha256(char *buf32bytes, const void *, size_t);
extern int scan_dirs();
extern int link_dirs();

extern int __attribute__((format(printf, 1, 2)))
err_print(const char *format, ...);
extern int __attribute__((format(printf, 1, 2)))
err_perror(const char *format, ...);

/* log line */
extern int __attribute__((format(printf, 1, 2)))
log_debug(const char *format, ...);
extern int __attribute__((format(printf, 1, 2)))
log_print(const char *format, ...);
extern int __attribute__((format(printf, 1, 2)))
log_verbose(const char *format, ...);

/* log no-newline */
extern int __attribute__((format(printf, 1, 2)))
logn_debug(const char *format, ...);
extern int __attribute__((format(printf, 1, 2)))
logn_print(const char *format, ...);
extern int __attribute__((format(printf, 1, 2)))
logn_verbose(const char *format, ...);


#endif /* _LOCAL_H_ */

/*
  Local variables: ***
  mode: c ***
  coding: iso-8859-15-unix ***
  c-basic-offset: 4 ***
  indent-tabs-mode: nil ***
  End: ***
*/
