#include "local.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <errno.h>

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/resource.h>

/*

  Copyright (c) 2024 Juha Railoma

  Permission to use, copy, modify, and/or distribute this software for
  any purpose with or without fee is hereby granted.

  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
  WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
  OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
  FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
  OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

*/


static void
__main_print_version_and_exit() {

    fprintf(stdout, "%s %s\n", app_progname, VERSION);
    exit(0);
}

static void
__main_print_help_and_exit() {

    fprintf(stdout, "Usage: %s [OPTIONS] [directory]...\n"
            "Hardlink identical files in given directories\n"
            "\n"
            "OPTIONS\n"
            "\n"
            "  -h, --help                 Print this message and exit\n"
            "  -v, --verbose              Print extra messages and report hardlinked files.\n"
            "  -q, --quiet                Don't tell anything. Even errors.\n"
            "  -d, --debug\n"
            "\n"
            "  -n, --dry-run              Don't change or link anything, just report\n"
            "                             what would be done.\n"
            "  -c, --content-only         For candidates compare only file contents.\n"
            "  -k, --keep-going           Keep going despite errors.\n"
            "  -l, --max-links=LINKS      Allow maximum LINKS hardlinks created\n"
            "                             per single file.\n"
            "  -V, --version              Print the version number and exit.\n"
            "\n"
            "Program scans throught given directories and seeks identical files.\n"
            "\n"
            "Identical files are files having same owner, same permissions and \nsame SHA256 sum of their contents, except if --contents-only option \nis given. In this case only file contents are considered.\n"
            "\n"
            "After discovering, program tries to replace all identical files with \nhardlinks to minimal number of physical files. If there would be more \nlinks than is allowed by filesystem design or amount is specifically \nlimited by --max-links, program leaves enough duplicates to \naccommodate required amount of links.\n"
            "\n"
            "Program terminates on first encountered error, except if \noption --keep-going is specified.\n"
            "\n"
            "Option --dry-run is suitable for checking what would be done. \nIt's good idea to give also --verbose to see what files would be \nhardlinked. Dry run mode does not create anything, so for example \nerrors related to permissions will not be reported.\n"
            "\n"
            "Hardlinked files are practically same file with different names.\n"
            "If single file is modified, all hardlinked files are modified,\n"
            "so beware running command without fully understanding what you\n"
            "are doing.\n"
            "\n", app_progname);

    exit(0);
}

static char *
__main_trim_path(char *path) {

    char *cps, *cpe;

    /* path is in stack */
    for(cps = cpe = path; *cpe; cps++, cpe++) {
        if ((cpe[0] == '/') && (cpe[1] == '/'))
            while((cpe[1] == '/'))
                cpe++;
        *cps = *cpe;
    }

    *cps = 0;
    if ((cps > path) && (cps[-1] == '/'))
        cps[-1] = 0;

    return path;
}

static int
__main_precheck_dirs() {

    int i;
    struct stat st;

    for(i = 0; i < app_dirs_count; i++) {

        long l;

        /* rtrim trailing and consecutive slashes */
        app_dirs[i] = __main_trim_path(app_dirs[i]);
        app_dirs[i] = obstack_strdup(obs, app_dirs[i]);

        if (stat(app_dirs[i], &st) < 0)
            return err_perror("Stat failed on directory %s", app_dirs[i]);

        if ((st.st_mode & S_IFMT) != S_IFDIR)
            return err_print("Argument %s is not directory", app_dirs[i]);

        /* nlinks-max */
        if ((l = pathconf(app_dirs[i], _PC_LINK_MAX)) < 0)
            return err_print("pathconf(_PC_LINK_MAX) failed on %s",
                             app_dirs[i]);

        if (l > 0x7fffffff) {
            /* struct stat.st_nlink is 64bit, to save memory let's
             * truncate it to 32bit since ext4 limit is anyway just
             * 65000 */
            l = 0x7fffffff;
        }

        if (app_nlinks_max && (app_nlinks_max != l)) {
            return err_print("_PC_LINK_MAX differs between directories, "
                             "got %li while last was %i",
                             l, app_nlinks_max);
        } else {
            app_nlinks_max = (int)l;
        }

        /* path-max */
        if ((l = pathconf(app_dirs[i], _PC_PATH_MAX)) < 0)
            return err_print("pathconf(_PC_PATH_MAX) failed on %s",
                             app_dirs[i]);

        if (l > app_path_max)
            app_path_max = l;
    }

    /* sanity-check */
    if (app_path_max < 512)
        return err_print("Error, cant belive _PC_PATH_MAX %i below 512",
                         app_path_max);

    return 0;
}

static int
__main_parse_args(int argc, char **argv) {

    static const struct option longopts[] = {
        { "help", no_argument, 0, 2 },
        { "verbose", no_argument, 0, 3 },
        { "quiet", no_argument, 0, 4 },
        { "debug", no_argument, 0, 5 },
        { "content-only", no_argument, 0, 6 },
        { "dry-run", no_argument, 0, 7 },
        { "keep-going", no_argument, 0, 8 },
        { "max-links", required_argument, 0, 9 },
        { "version", no_argument, 0, 10 },
        { NULL, 0, 0, 0 }
    };

    /* progname */
    app_progname = argv[0];
    if (strrchr(app_progname, (int)'/'))
        app_progname = strrchr(app_progname, (int)'/') + 1;

    for(;;) {

        int c = getopt_long(argc, argv, "-hvqdcnkVl:", longopts, NULL);

        if (c == -1)
            break;

        switch(c) {
        case 1: /* path-entry */
            app_dirs_count++;

            if (!optarg || !*optarg) {
                errno = EINVAL;
                return err_print("Blank directory path given");
            }

            obstack_ptr_grow(obs, n_stackdup(optarg));
            break;

        case 'h': /* --help */
        case 2:
            __main_print_help_and_exit();
            break;


        case 'v': /* --verbose */
        case 3:
            app_quiet = 0;
            app_verbose = 1;
            break;

        case 'q': /* --quiet */
        case 4:
            app_quiet = 1;
            app_verbose++;
            break;

        case 'd':
        case 5: /* --debug */
            app_quiet = 0;
            app_debug = 1;
            app_verbose = 1;
            break;

        case 'c':
        case 6: /* --content-only */
            app_content_only = 1;
            break;

        case 'n': /* --dry-run */
        case 7:
            app_dry_run = 1;
            break;

        case 'k': /* --keep-going */
        case 8:
            app_keep_going = 1;
            break;

        case 'l': /* --max-links */
        case 9:
            if (optarg && !(app_nlinks_limit = atoi(optarg)))
                app_nlinks_limit = -1;
            break;

        case 'V':
        case 10:
            __main_print_version_and_exit();
            break;
        }
    }

    /* construct and precheck dirs */
    obstack_ptr_grow(obs, NULL);
    app_dirs = obstack_finish(obs);
    if (!app_dirs_count)
        return err_print("Error, no directories given");

    if (__main_precheck_dirs() < 0)
        return -1;

    /* sanity-check */
    if (!app_nlinks_limit || (app_nlinks_limit > app_nlinks_max))
        app_nlinks_limit = app_nlinks_max;
    if (app_nlinks_limit < 2)
        return err_print("No point to hardlink, if nlinks-limit < 2");

    return 0;
}

static int
__main_defaults() {

    struct rlimit rl;
    static struct obstack __obs = { 0 };

    /* init obstacks and set large allocations to force malloc do
     * mmapped allocations */
    obs = &__obs;
    obstack_init(obs);

    /* it's virtual mem, every page is allocated only if actually
     * needed. */
    obstack_chunk_size(obs) = 0x1000000; /* 16MiB */

    /* syslimits */
    if (getrlimit(RLIMIT_NOFILE, &rl) < 0)
        return err_perror("Get max fd count failed");

    if (rl.rlim_cur < 256)
        return err_print("Can't belive max allowed fds below 256");

    app_fds_max = rl.rlim_cur;
    app_fds_safe = rl.rlim_cur - 128;

    return 0;
}

int main(int argc, char **argv) {

    if (!__main_defaults() &&
        !__main_parse_args(argc, argv) &&
        !scan_dirs() &&
        !link_dirs())
        return 0;

    return -1;
}


/*
  Local variables: ***
  mode: c ***
  coding: iso-8859-15-unix ***
  c-basic-offset: 4 ***
  indent-tabs-mode: nil ***
  End: ***
*/
