#include "local.h"

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <alloca.h>
#include <errno.h>

#include <dirent.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>

/*

  Copyright (c) 2024 Juha Railoma

  Permission to use, copy, modify, and/or distribute this software for
  any purpose with or without fee is hereby granted.

  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
  WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
  OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
  FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
  DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
  OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

*/

static int
__scan_queue_file(struct z_file_t *fe, off_t size) {

    struct z_sizes_t *sp, *nsp;

    stat_origin_files++;
    stat_origin_bytes += size;

    /* first entry */
    if (!(sp = app_sizes[size % Z_SIZES_HASH])) {
        app_sizes[size % Z_SIZES_HASH] =
            sp = obstack_alloc(obs, sizeof(*sp));

        fe->next = NULL;
        sp->next = NULL;
        sp->size = size;
        sp->files = fe;
        sp->count = 1;

        app_sizes_count++;
        return 0;
    }

    /* else queue */
    for(; sp; sp = sp->next) {

        if (sp->next && (sp->next->size >= size))
            continue;

        if (sp->size == size) {
            fe->next = sp->files;
            sp->files = fe;
            sp->count++;
            return 0;
        }

        /* queue new size */
        nsp = obstack_calloc(obs, sizeof(*nsp));
        nsp->size = size;
        nsp->files = fe;
        nsp->count = 1;
        nsp->next = sp->next;
        sp->next = nsp;

        app_sizes_count++;
        return 0;
    }

    /* never here */
    return -1;
}

static int
__scan_add_file(const char *parent, const char *name,
                struct stat *st) {

    struct z_file_t *fe;

    /* skip 0-byte files */
    if (st->st_size == 0)
        return 0;

    fe = obstack_alloc(obs, sizeof(*fe));
    fe->uid = st->st_uid;
    fe->gid = st->st_gid;
    fe->ino = st->st_ino;
    fe->mode = st->st_mode;
    fe->nlink = st->st_nlink;

    fe->parent = parent; /* parent is obstacked const */
    fe->name = obstack_strdup(obs, name);
    return __scan_queue_file(fe, st->st_size);
}

static int
__scan_dir(const char *parent, int depth) {

    DIR *dir;
    char *filepos;
    const char *slashparent;
    struct stat st;
    struct dirent *dent;
    struct z_slist_t *de, *dlist = NULL;

    /* statics */
    static dev_t __scan_device = 0;
    static char fullpath[8192 + 128];

    if (app_debug) {
        for(int i = 0; i < depth; i++)
            logn_debug("  ");
        log_debug("  %s", parent);
    }

    if (strlen(parent) > 4096)
        return (errno = EINVAL),
            err_perror("Path length over 4096 bytes");

    /* PATH_MAX is difficult. very probable it's 4096 but actual
     * length of path might be anything. we shortcut by assuming 4096
     * for parent and 4096 for name. if fullpath -1 is something than
     * zero, something is messed up and we break out

     * (I JUST HATE calculating exact bytes - when there's actually
     * gigabytes on memory available) */

    /* note: extra trailing slashes has been rtrimmed in main, but we
     * store parent-name (once) with trailing slash */
    strncpy(fullpath, parent, 4096);
    filepos = fullpath + strlen(fullpath);
    *filepos++ = '/';

    slashparent = obstack_strdup(obs, fullpath);

    if (!(dir = opendir(parent))) {
        stat_dir_open_failed++;
        err_perror("   opendir %s failed", parent);
        return 0;
    }

    while((dent = readdir(dir))) {

        /* skip . and .. */
        if ((dent->d_name[0] == '.') &&
            (!dent->d_name[1] ||
             (dent->d_name[1] == '.' && !dent->d_name[2])))
                continue;

        switch(dent->d_type) {
        case DT_REG:
            stat_file_count++;
            strcpy(filepos, dent->d_name);

            /* stat to get current size */
            if (lstat(fullpath, &st) < 0) {
                stat_file_stat_failed++;
                err_perror("   stat file %s failed", fullpath);
                break;
            }

            __scan_add_file(slashparent, dent->d_name, &st);
            break;

        case DT_DIR:

            stat_dir_count++;
            if (strlen(dent->d_name) > 4096)
                return (errno = EINVAL),
                    err_perror("Filename absurdity error");

            strcpy(filepos, dent->d_name);

            /* check target directory on same device */
            if (lstat(fullpath, &st) < 0) {
                stat_dir_stat_failed++;
                err_perror(" stat directory %s failed", fullpath);
                break;
            }
            if (__scan_device) {
                if (__scan_device != st.st_dev) {
                    log_print(" %s on different filesystem, "
                              "won't descend", fullpath);
                    break;
                }
            } else {
                __scan_device = st.st_dev;
            }

            /* if yes, collect directory to be processed later */
            de = n_stmalloca(de);
            de->next = dlist;
            de->val = obstack_strdup(obs, fullpath);
            dlist = de;
            break;

        case DT_BLK: stat_blk_count++; break;
        case DT_CHR: stat_chr_count++; break;
        case DT_LNK: stat_lnk_count++; break;
        case DT_FIFO: stat_fifo_count++; break;
        case DT_SOCK: stat_sock_count++; break;
        default:
            stat_unknown_count++; break;
        }
    }

    /* process collected directories after closedir to avoid fd leak */
    closedir(dir);

    while(dlist) {

        if (__scan_dir((const char *)dlist->val, (depth + 1)) < 0)
            return -1;
        dlist = dlist->next;
    }

    /* success */
    return 0;
}

int
scan_dirs() {

    int i;
    int64_t count_dirs;
    int64_t count_files;

    /* and collect */
    for(i = 0; i < app_dirs_count; i++) {
        log_print("Scanning directory %s...", app_dirs[i]);

        count_dirs = stat_dir_count;
        count_files = stat_file_count;
        if (__scan_dir(app_dirs[i], 0) < 0) {
            return -1;
        }

        log_verbose(" total: %zi dirs, %zi files",
                    (stat_dir_count - count_dirs),
                    (stat_file_count - count_files));
    }

    return 0;
}

/*
  Local variables: ***
  mode: c ***
  coding: iso-8859-15-unix ***
  c-basic-offset: 4 ***
  indent-tabs-mode: nil ***
  End: ***
*/
