#include "local.h"

#include <string.h>
#include <stdint.h>
#include <stdarg.h>

/* lib-sha256.c */
struct sha256_ctx_t {
    uint32_t datalen;
    uint32_t bitlen[2];
    uint32_t state[8];
    unsigned char data[64];
};

/* sha-makrot */
#define DA(a, b, c) { if (a > (0xffffffff - (c))) ++b; a += c; }
#define SHR(b, w)   ((w) >> (b))
#define ROL(b, w)  (((w) << (b)) | ((w) >> (32 - (b))))
#define ROR(b, w)  (((w) >> (b)) | ((w) << (32 - (b))))
#define CH(x, y, z)  (((x) & (y)) ^ (~(x) & (z)))
#define MAJ(x, y, z) (((x) & (y)) ^ ((x) & (z)) ^ ((y) & (z)))
#define SI0(w) (ROR(2, w) ^ ROR(13, w) ^ ROR(22, w))
#define SI1(w) (ROR(6, w) ^ ROR(11, w) ^ ROR(25, w))
#define si0(w) (ROR(7, w) ^ ROR(18, w) ^ SHR(3,  w))
#define si1(w) (ROR(17,w) ^ ROR(19, w) ^ SHR(10, w))

/* reduced from other projects, origin RFC6234 */
static inline void
__sha256_transform(struct sha256_ctx_t *ctx) {

    int32_t t1, t4;
    uint32_t temp1, temp2;
    uint32_t A, B, C, D, E, F, G, H, M[64];
    unsigned char *data = ctx->data;

    static const uint32_t K[64] = {
	0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5,
	0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
	0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
	0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
	0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc,
	0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
	0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7,
	0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
	0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
	0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
	0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3,
	0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
	0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5,
	0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
	0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
	0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
    };

    for(t1 = t4 = 0; t1 < 16; t1++, t4 += 4) {
	M[t1] = (data[t4] << 24) |
	    (data[t4 + 1] << 16) |
	    (data[t4 + 2] << 8) |
	    (data[t4 + 3]);
    }

    for(t1 = 16; t1 < 64; t1++) {
	M[t1] = si1(M[t1 - 2]) + M[t1 - 7] +
	    si0(M[t1 - 15]) + M[t1 - 16];
    }

    A = ctx->state[0];
    B = ctx->state[1];
    C = ctx->state[2];
    D = ctx->state[3];
    E = ctx->state[4];
    F = ctx->state[5];
    G = ctx->state[6];
    H = ctx->state[7];

    for (t1 = 0; t1 < 64; t1++) {
	temp1 = H + SI1(E) + CH(E, F, G) + K[t1] + M[t1];
	temp2 = SI0(A) + MAJ(A, B, C);
	H = G;
	G = F;
	F = E;
	E = D + temp1;
	D = C;
	C = B;
	B = A;
	A = temp1 + temp2;
    }

    ctx->state[0] += A;
    ctx->state[1] += B;
    ctx->state[2] += C;
    ctx->state[3] += D;
    ctx->state[4] += E;
    ctx->state[5] += F;
    ctx->state[6] += G;
    ctx->state[7] += H;
}

static inline void
__sha256_update(struct sha256_ctx_t *ctx, const void *vptr, size_t len) {

    uint32_t i;
    const unsigned char *data =
	(const unsigned char *)vptr;

    for(i = 0; i < len; ++i) {

	ctx->data[ctx->datalen++] = data[i];
	if (ctx->datalen == 64) {
	    __sha256_transform(ctx);
	    DA(ctx->bitlen[0], ctx->bitlen[1], 512);
	    ctx->datalen = 0;
	}
    }
}

static inline const char *
__sha256_finalize(struct sha256_ctx_t *ctx, char *bptr) {

    uint32_t i;
    unsigned char *result = (unsigned char *)bptr;

    i = ctx->datalen;
    ctx->data[i++] = 0x80;
    if (ctx->datalen < 56) {
	while(i < 56) ctx->data[i++] = 0;
    } else {
	while(i < 64) ctx->data[i++] = 0;
	__sha256_transform(ctx);
	bzero(ctx->data, 56);
    }

    DA(ctx->bitlen[0], ctx->bitlen[1], ctx->datalen * 8);
    ctx->data[63] = ctx->bitlen[0];
    ctx->data[62] = ctx->bitlen[0] >> 8;
    ctx->data[61] = ctx->bitlen[0] >> 16;
    ctx->data[60] = ctx->bitlen[0] >> 24;
    ctx->data[59] = ctx->bitlen[1];
    ctx->data[58] = ctx->bitlen[1] >> 8;
    ctx->data[57] = ctx->bitlen[1] >> 16;
    ctx->data[56] = ctx->bitlen[1] >> 24;
    __sha256_transform(ctx);

    /* little-endian */
    for (i=0; i < 4; ++i) {
	result[i]    = (ctx->state[0] >> (24-i*8)) & 0x000000ff;
	result[i + 4]  = (ctx->state[1] >> (24-i*8)) & 0x000000ff;
	result[i + 8]  = (ctx->state[2] >> (24-i*8)) & 0x000000ff;
	result[i + 12] = (ctx->state[3] >> (24-i*8)) & 0x000000ff;
	result[i + 16] = (ctx->state[4] >> (24-i*8)) & 0x000000ff;
	result[i + 20] = (ctx->state[5] >> (24-i*8)) & 0x000000ff;
	result[i + 24] = (ctx->state[6] >> (24-i*8)) & 0x000000ff;
	result[i + 28] = (ctx->state[7] >> (24-i*8)) & 0x000000ff;
    }

    return bptr;
}

const char *
sha256(char *buf32bytes, const void *vptr, size_t datalen) {

    const char *data = (const char *)vptr;
    struct sha256_ctx_t ctx = {
        .datalen = 0,
        .bitlen = { 0, 0 },
        .state = {
            0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a,
            0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19
        }
    };

    __sha256_update(&ctx, (const unsigned char *)data, datalen);
    return __sha256_finalize(&ctx, buf32bytes);
}

/*
  Local variables: ***
  mode: c ***
  coding: iso-8859-15-unix ***
  c-basic-offset: 4 ***
  indent-tabs-mode: nil ***
  End: ***
*/
